# -*- encoding: utf-8 -*-
"""
keri.kli.common.terming module

"""


class Colors:
    """ Terminal colors """
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class Symbols:
    """ Terminal Symbols """
    CHECKMARK = '\u2714'
    FAILED = '\u2718'
